﻿using System;
using System.Windows.Forms;

namespace WindowsFormsApp1
{
    // https://stackoverflow.com/questions/14584280/c-sharp-get-mouse-handle-getrawinputdeviceinfo
    public partial class Form1 : Form, IMessageFilter
    {
        public Form1()
        {
            InitializeComponent();
            Application.AddMessageFilter(this);
        }

        public bool PreFilterMessage(ref Message m)
        {
            if (m.Msg == Win32Interop.WM_INPUT)
            {
                RawInput raw = Win32Interop.GetDeviceID(m);

                string source = (m.WParam.ToInt32() == Win32Interop.RIM_INPUT) ? "this" : "others";

                switch (raw.Header.Type)
                {
                    case RawInputType.Keyboard:
                        System.Diagnostics.Debug.WriteLine(
                            $"[{source}] [keyboard] Device ID : {raw.Header.Device}, VKey : {raw.Keyboard.VKey}, Code : {raw.Keyboard.MakeCode} Msg: {raw.Keyboard.Message}");
                        break;

                    case RawInputType.Mouse:
                        System.Diagnostics.Debug.WriteLine(
                            $"[{source}] [mouse] Device ID : {raw.Header.Device}, X : {raw.Mouse.LastX}, Y : {raw.Mouse.LastY}");
                        break;
                }
            }

            return false;
        }

        protected override void OnFormClosing(FormClosingEventArgs e)
        {
            Application.RemoveMessageFilter(this);
            base.OnFormClosing(e);
        }

        private unsafe void Form1_Load(object sender, EventArgs e)
        {
            RAWINPUTDEVICE [] Rid = new RAWINPUTDEVICE[2];

            Rid[0].UsagePage = HIDUsagePage.Generic;
            Rid[0].Usage = HIDUsage.Mouse;
            Rid[0].Flags = RawInputDeviceFlags.InputSink;
            Rid[0].WindowHandle = this.Handle;

            Rid[1].UsagePage = HIDUsagePage.Generic;
            Rid[1].Usage = HIDUsage.Keyboard;
            Rid[1].Flags = RawInputDeviceFlags.InputSink;
            Rid[1].WindowHandle = this.Handle;

            if (Win32Interop.RegisterRawInputDevices(Rid, 2, sizeof(RAWINPUTDEVICE)) == false)
            {
                MessageBox.Show("Failed to register");
            }
        }
    }
}
